/*
 * aconvert.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Martin Zak
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>
#include <factory/XmlDataFactory.hpp>
#include <registry/Registry.h>

int main ( int argc, char * argv[] ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Introspection binary", ' ', "0.01" );

		TCLAP::SwitchArg dataTypes ( "d", "datatypes", "Data types", false );
		cmd.add ( dataTypes );

		TCLAP::SwitchArg algorithms ( "a", "algorithms", "Algorithms", false );
		cmd.add ( algorithms );

		TCLAP::SwitchArg overloads ( "o", "overloads", "Overloads", false );
		cmd.add ( overloads );

		TCLAP::SwitchArg castFrom ( "f", "cast_from", "Casting from type", false );
		cmd.add ( castFrom );

		TCLAP::SwitchArg castTo ( "t", "cast_to", "Casting to type", false );
		cmd.add ( castTo );

		TCLAP::SwitchArg summary ( "s", "sumary", "Sumary", false );
		cmd.add ( summary );

		TCLAP::UnlabeledValueArg < std::string > argument ( "argument", "Type name for casts, algorithm name for overloads, group name for algorithms or datatypes", false, "", "string" );
		cmd.add ( argument );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if ( verbose.isSet ( ) )
			common::GlobalData::verbose = true;

		if ( measure.isSet ( ) )
			common::GlobalData::measure = true;

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Output write", measurements::Type::AUXILIARY );

		cli::Environment environment;
		environment.setBinding ( "argument", argument.getValue ( ) );

		if ( dataTypes.isSet ( ) ) {
			cli::Parser parser ( cli::Lexer ( "introspect datatypes #argument" ) );
			parser.parse ( )->run ( environment );
		} else if ( algorithms.isSet ( ) ) {
			cli::Parser parser ( cli::Lexer ( "introspect algorithms #argument" ) );
			parser.parse ( )->run ( environment );
		} else if ( overloads.isSet ( ) ) {
			cli::Parser parser ( cli::Lexer ( "introspect overloads #argument" ) );
			parser.parse ( )->run ( environment );
		} else if ( castFrom.isSet ( ) ) {
			cli::Parser parser ( cli::Lexer ( "introspect casts :from #argument" ) );
			parser.parse ( )->run ( environment );
		} else if ( castTo.isSet ( ) ) {
			cli::Parser parser ( cli::Lexer ( "introspect casts :to #argument" ) );
			parser.parse ( )->run ( environment );
		} else if ( summary.isSet ( ) ) {
			ext::deque < sax::Token > tokens;

			tokens.emplace_back("Summary", sax::Token::TokenType::START_ELEMENT);
			tokens.emplace_back("Casts", sax::Token::TokenType::START_ELEMENT);
			for ( const ext::pair < std::string, std::string > & cast : abstraction::Registry::listCasts ( ) ) {
				tokens.emplace_back("Cast", sax::Token::TokenType::START_ELEMENT);
				tokens.emplace_back("From", sax::Token::TokenType::START_ELEMENT);
				tokens.emplace_back(cast.second, sax::Token::TokenType::CHARACTER);
				tokens.emplace_back("From", sax::Token::TokenType::END_ELEMENT);
				tokens.emplace_back("To", sax::Token::TokenType::START_ELEMENT);
				tokens.emplace_back(cast.first, sax::Token::TokenType::CHARACTER);
				tokens.emplace_back("To", sax::Token::TokenType::END_ELEMENT);
				tokens.emplace_back("Cast", sax::Token::TokenType::END_ELEMENT);
			}
			tokens.emplace_back("Casts", sax::Token::TokenType::END_ELEMENT);
			tokens.emplace_back("Algorithms", sax::Token::TokenType::START_ELEMENT);
			for ( const std::string & algo : abstraction::Registry::listAlgorithms ( ) ) {
				tokens.emplace_back("Algorithm", sax::Token::TokenType::START_ELEMENT);
				tokens.emplace_back("Name", sax::Token::TokenType::START_ELEMENT);
				tokens.emplace_back(algo, sax::Token::TokenType::CHARACTER);
				tokens.emplace_back("Name", sax::Token::TokenType::END_ELEMENT);
				tokens.emplace_back("Overloads", sax::Token::TokenType::START_ELEMENT);
				for ( const ext::tuple < abstraction::AlgorithmCategories::AlgorithmCategory, ext::pair < std::string, ext::set < abstraction::ParamQualifiers::ParamQualifier > >, ext::vector < ext::tuple < std::string, ext::set < abstraction::ParamQualifiers::ParamQualifier >, std::string > > > & over : abstraction::Registry::listOverloads ( algo ) ) {
					tokens.emplace_back("Overload", sax::Token::TokenType::START_ELEMENT);
					tokens.emplace_back("Category", sax::Token::TokenType::START_ELEMENT);
					tokens.emplace_back(abstraction::AlgorithmCategories::toString ( std::get < 0 > ( over ) ), sax::Token::TokenType::CHARACTER);
					tokens.emplace_back("Category", sax::Token::TokenType::END_ELEMENT);
					tokens.emplace_back("Returns", sax::Token::TokenType::START_ELEMENT);
					tokens.emplace_back("Type", sax::Token::TokenType::START_ELEMENT);
					tokens.emplace_back(std::get < 1 > ( over ).first, sax::Token::TokenType::CHARACTER);
					tokens.emplace_back("Type", sax::Token::TokenType::END_ELEMENT);
					tokens.emplace_back("Qualifications", sax::Token::TokenType::START_ELEMENT);
					for ( abstraction::ParamQualifiers::ParamQualifier qualifier : std::get < 1 > ( over ).second ) {
						tokens.emplace_back ( "Qualifier", sax::Token::TokenType::START_ELEMENT);
						switch ( qualifier ) {
						case abstraction::ParamQualifiers::ParamQualifier::CONST:
							tokens.emplace_back("const", sax::Token::TokenType::CHARACTER);
							break;
						case abstraction::ParamQualifiers::ParamQualifier::LREF:
							tokens.emplace_back("lref", sax::Token::TokenType::CHARACTER);
							break;
						case abstraction::ParamQualifiers::ParamQualifier::RREF:
							tokens.emplace_back("rref", sax::Token::TokenType::CHARACTER);
							break;
						}
						tokens.emplace_back ( "Qualifier", sax::Token::TokenType::END_ELEMENT);
					}
					tokens.emplace_back("Qualifications", sax::Token::TokenType::END_ELEMENT);
					tokens.emplace_back("Returns", sax::Token::TokenType::END_ELEMENT);
					tokens.emplace_back("Params", sax::Token::TokenType::START_ELEMENT);
					for ( const ext::tuple < std::string, ext::set < abstraction::ParamQualifiers::ParamQualifier >, std::string > & param : std::get < 2 > ( over ) ) {
						tokens.emplace_back("Param", sax::Token::TokenType::START_ELEMENT);
						tokens.emplace_back("Type", sax::Token::TokenType::START_ELEMENT);
						tokens.emplace_back(std::get < 0 > ( param ), sax::Token::TokenType::CHARACTER);
						tokens.emplace_back("Type", sax::Token::TokenType::END_ELEMENT);
						tokens.emplace_back("Qualifications", sax::Token::TokenType::START_ELEMENT);
						for ( abstraction::ParamQualifiers::ParamQualifier qualifier : std::get < 1 > ( param ) ) {
							tokens.emplace_back ( "Qualifier", sax::Token::TokenType::START_ELEMENT);
							switch ( qualifier ) {
							case abstraction::ParamQualifiers::ParamQualifier::CONST:
								tokens.emplace_back("const", sax::Token::TokenType::CHARACTER);
								break;
							case abstraction::ParamQualifiers::ParamQualifier::LREF:
								tokens.emplace_back("lref", sax::Token::TokenType::CHARACTER);
								break;
							case abstraction::ParamQualifiers::ParamQualifier::RREF:
								tokens.emplace_back("rref", sax::Token::TokenType::CHARACTER);
								break;
							}
							tokens.emplace_back ( "Qualifier", sax::Token::TokenType::END_ELEMENT);
						}
						tokens.emplace_back("Qualifications", sax::Token::TokenType::END_ELEMENT);
						tokens.emplace_back("Name", sax::Token::TokenType::START_ELEMENT);
						tokens.emplace_back(std::get < 2 > ( param ), sax::Token::TokenType::CHARACTER);
						tokens.emplace_back("Name", sax::Token::TokenType::END_ELEMENT);
						tokens.emplace_back("Param", sax::Token::TokenType::END_ELEMENT);
					}
					tokens.emplace_back("Params", sax::Token::TokenType::END_ELEMENT);
					tokens.emplace_back("Overload", sax::Token::TokenType::END_ELEMENT);
				}
				tokens.emplace_back("Overloads", sax::Token::TokenType::END_ELEMENT);
				tokens.emplace_back("Algorithm", sax::Token::TokenType::END_ELEMENT);
			}
			tokens.emplace_back("Algorithms", sax::Token::TokenType::END_ELEMENT);
			tokens.emplace_back("Summary", sax::Token::TokenType::END_ELEMENT);

			sax::SaxComposeInterface::printStdout ( tokens );
		}

		measurements::end ( );
		measurements::end ( );

		if ( measure.getValue ( ) ) common::Streams::measure << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		common::Streams::err << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		common::Streams::err << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		common::Streams::err << "Unknown exception caught." << std::endl;
		return 127;
	}
}
