/*
 * LeftRecursionRemover.h
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#ifndef LEFT_RECURSION_REMOVER_H_
#define LEFT_RECURSION_REMOVER_H_

#include <algorithm>

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <grammar/convert/ToGrammarRightRG.h>
#include <exception/CommonException.h>

#include <vector>
#include <common/createUnique.hpp>

namespace grammar {

namespace simplify {

class LeftRecursionRemover {
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > directLeftRecursionRemoveAsOrder ( const grammar::EpsilonFreeCFG < SymbolType > & origGrammar );

	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > assignAsOrder ( const grammar::EpsilonFreeCFG < SymbolType > & origGrammar, unsigned i, const ext::set< SymbolType >& origNonterminals );
public:
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::EpsilonFreeCFG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::CNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > remove( const grammar::GNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightRG < SymbolType > remove( const grammar::RightRG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightLG < SymbolType > remove( const grammar::RightLG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightRG < SymbolType > remove( const grammar::LeftRG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightLG < SymbolType > remove( const grammar::LeftLG < SymbolType > & grammar );
};

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > LeftRecursionRemover::directLeftRecursionRemoveAsOrder(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar) {
	grammar::EpsilonFreeCFG < SymbolType > res(origGrammar.getInitialSymbol());
	res.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	res.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	res.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	for(const auto& nonterminal : origGrammar.getNonterminalAlphabet()) {
		if(origGrammar.getRules().find(nonterminal) == origGrammar.getRules().end()) continue;

		if(std::any_of(origGrammar.getRules().find(nonterminal)->second.begin(), origGrammar.getRules().find(nonterminal)->second.end(), [&](const ext::vector<SymbolType>& singleRHS) {
				return singleRHS[0] == nonterminal; // is there a direct left recursion?
			} ) && std::all_of(origGrammar.getRules().find(nonterminal)->second.begin(), origGrammar.getRules().find(nonterminal)->second.end(), [&](const ext::vector<SymbolType>& singleRHS) {
				return origGrammar.getTerminalAlphabet().count(singleRHS[0]) || singleRHS[0] >= nonterminal; // only remove left recursion when all nonterminals are bigger than the left hand side
			})) {
			SymbolType primed = common::createUnique(nonterminal, res.getTerminalAlphabet(), res.getNonterminalAlphabet());
			res.addNonterminalSymbol(primed);
			for(const ext::vector<SymbolType>& singleRHS : origGrammar.getRules().find(nonterminal)->second) { // do the removal
				if(singleRHS[0] == nonterminal) { // A -> A alpha
					ext::vector<SymbolType> tmpRHS(singleRHS.begin() + 1, singleRHS.end());

					res.addRule(primed, tmpRHS); // A' -> alpha

					tmpRHS.push_back(primed);
					res.addRule(primed, tmpRHS); // A' -> alpha A'
				} else { // a -> beta
					ext::vector<SymbolType> tmpRHS(singleRHS);

					res.addRule(nonterminal, tmpRHS); // A -> beta

					tmpRHS.push_back(primed);
					res.addRule(nonterminal, tmpRHS); // A -> beta A'
				}
			}
		} else {
			for(const ext::vector<SymbolType>& singleRHS : origGrammar.getRules().find(nonterminal)->second) {
				res.addRule(nonterminal, singleRHS);
			}
		}
	}
	return res;
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > LeftRecursionRemover::assignAsOrder(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar, unsigned i, const ext::set<SymbolType>& origNonterminals) {
	grammar::EpsilonFreeCFG < SymbolType > res(origGrammar.getInitialSymbol());
	res.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	res.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	res.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	for(const SymbolType& lhs : origGrammar.getNonterminalAlphabet()) {
		if(i > 0) {
			if(origGrammar.getRules().find(lhs) == origGrammar.getRules().end()) continue;
			for(const ext::vector<SymbolType>& rule : origGrammar.getRules().find(lhs)->second) {
				res.addRule(lhs, rule);
			}

			i--;
			continue; // substitue only in i-th up to n-th nonterminals
		}
		if(origGrammar.getRules().find(lhs) == origGrammar.getRules().end()) continue;
		if(!origNonterminals.count(lhs)) { // do not subsitute in nonoriginal nonterminals
			for(const ext::vector<SymbolType>& rule : origGrammar.getRules().find(lhs)->second) {
				res.addRule(lhs, rule);
			}
			continue;
		}

		const ext::set<ext::vector<SymbolType>>& rule = origGrammar.getRules().find(lhs)->second;

		for(const ext::vector<SymbolType>& singleRHS : rule) {
			if(res.getTerminalAlphabet().count(singleRHS[0])) { //do not substitute terminals
				res.addRule(lhs, singleRHS);
				continue;
			}
			const SymbolType& secondLHS = singleRHS[0];
			if(secondLHS >= lhs) { // substitute only by 0th up to i-th nonterminals right hand sides
				res.addRule(lhs, singleRHS);
				continue;
			}
			if(origGrammar.getRules().find(secondLHS) == origGrammar.getRules().end()) { //is there any right hand side to substitue with?
				//if not well this rule does not generate anything anyway
				continue;
			}

			for(const ext::vector<SymbolType>& secondSingleRHS : origGrammar.getRules().find(secondLHS)->second) { // do the substitution
				ext::vector<SymbolType> newRHS(secondSingleRHS);
				newRHS.insert(newRHS.end(), singleRHS.begin() + 1, singleRHS.end());
				res.addRule(lhs, newRHS);
			}
		}
	}
	return res;
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > LeftRecursionRemover::remove(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar) {
	grammar::EpsilonFreeCFG < SymbolType > step = origGrammar;
	for(const SymbolType& nonterminal : step.getNonterminalAlphabet()) { // remove identities
		step.removeRule(nonterminal, ext::vector<SymbolType>{nonterminal});
	}
	unsigned i = 0;
	while(i < origGrammar.getNonterminalAlphabet().size()) {
		grammar::EpsilonFreeCFG < SymbolType > nextStep = assignAsOrder(directLeftRecursionRemoveAsOrder(step), i, origGrammar.getNonterminalAlphabet());

		if(step == nextStep) break;
		step = std::move(nextStep);
		i++;
	};

	return step;
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > LeftRecursionRemover::remove(const grammar::CNF < SymbolType > & origGrammar) {
	EpsilonFreeCFG < SymbolType > tmp(origGrammar.getInitialSymbol());
	tmp.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	tmp.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	tmp.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());
	for(const auto& rule : origGrammar.getRules()) {
		for(const auto& rhs : rule.second) {
			if(rhs.template is<SymbolType>()) {
				tmp.addRule(rule.first, {rhs.template get<SymbolType>()});
			} else {
				const auto& rhsPair = rhs.template get<ext::pair<SymbolType, SymbolType>>();
				tmp.addRule(rule.first, {rhsPair.first, rhsPair.second});
			}
		}
	}
	return remove(tmp);
}

template < class SymbolType >
grammar::GNF < SymbolType > LeftRecursionRemover::remove(const grammar::GNF < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::RightRG < SymbolType > LeftRecursionRemover::remove(const grammar::RightRG < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::RightLG < SymbolType > LeftRecursionRemover::remove(const grammar::RightLG < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::RightRG < SymbolType > LeftRecursionRemover::remove(const grammar::LeftRG < SymbolType > & origGrammar) {
	return convert::ToGrammarRightRG::convert(origGrammar);
}

template < class SymbolType >
grammar::RightLG < SymbolType > LeftRecursionRemover::remove(const grammar::LeftLG < SymbolType > & /* origGrammar */) {
	throw exception::CommonException("LeftRecursionRemover: Removing from LeftLG NYI"); // TODO
}

} /* namespace simplify */

} /* namespace grammar */

#endif /* LEFT_RECURSION_REMOVER_H_ */
