/*
 * SimpleRulesRemover.h
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#ifndef SIMPLE_RULES_REMOVER_H_
#define SIMPLE_RULES_REMOVER_H_

#include <algorithm>

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <grammar/properties/NonterminalUnitRuleCycle.h>

#include <grammar/Grammar.h>

namespace grammar {

namespace simplify {

class SimpleRulesRemover {
	template<class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static T removeEpsilonFree( const T & origGrammar );

	template<class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static T removeNonEpsilonFree( const T & origGrammar );

public:
	template < class SymbolType >
	static grammar::CFG < SymbolType > remove( const grammar::CFG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::EpsilonFreeCFG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > remove( const grammar::GNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::CNF < SymbolType > remove( const grammar::CNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::LG < SymbolType > remove( const grammar::LG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::LeftLG < SymbolType > remove( const grammar::LeftLG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::LeftRG < SymbolType > remove( const grammar::LeftRG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightLG < SymbolType > remove( const grammar::RightLG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightRG < SymbolType > remove( const grammar::RightRG < SymbolType > & grammar );
};

template < class T, class SymbolType >
T SimpleRulesRemover::removeEpsilonFree( const T & origGrammar ) {
	T grammar(origGrammar.getInitialSymbol());

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() )
		grammar.addNonterminalSymbol( symbol );

	for( const auto & symbol : origGrammar.getTerminalAlphabet() )
		grammar.addTerminalSymbol( symbol );

	auto origRules = origGrammar.getRawRules();

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() ) {
		ext::set<SymbolType> simpleRulesClosure = grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(origGrammar, symbol);
		for( const auto & closureSymbol : simpleRulesClosure ) {
			auto rules = origRules.find(closureSymbol);
			if(rules != origRules.end()) for( const auto& rawRule : rules->second ) {
				if(rawRule.size() == 0 || (rawRule.size() == 1 && !origGrammar.getNonterminalAlphabet().count(rawRule[0])) || rawRule.size() >= 2)
					grammar.addRawRule(symbol, rawRule);
			}
		}
	}

	grammar.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	return grammar;
}

template < class T, class SymbolType >
T SimpleRulesRemover::removeNonEpsilonFree( const T & origGrammar ) {
	T grammar(origGrammar.getInitialSymbol());

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() )
		grammar.addNonterminalSymbol( symbol );

	for( const auto & symbol : origGrammar.getTerminalAlphabet() )
		grammar.addTerminalSymbol( symbol );

	auto origRules = origGrammar.getRawRules();

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() ) {
		ext::set<SymbolType> simpleRulesClosure = grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(origGrammar, symbol);
		for( const auto & closureSymbol : simpleRulesClosure ) {
			auto rules = origRules.find(closureSymbol);
			if(rules != origRules.end()) for( const auto& rawRule : rules->second ) {
				if(rawRule.size() == 0 || (rawRule.size() == 1 && !origGrammar.getNonterminalAlphabet().count(rawRule[0])) || rawRule.size() >= 2)
					grammar.addRawRule(symbol, rawRule);
			}
		}
	}

	return grammar;
}

template < class SymbolType >
grammar::CFG < SymbolType > SimpleRulesRemover::remove(const grammar::CFG < SymbolType > & origGrammar) {
	return removeNonEpsilonFree(origGrammar);
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > SimpleRulesRemover::remove(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar) {
	return removeEpsilonFree(origGrammar);
}

template < class SymbolType >
grammar::CNF < SymbolType > SimpleRulesRemover::remove(const grammar::CNF < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::GNF < SymbolType > SimpleRulesRemover::remove(const grammar::GNF < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::LG < SymbolType > SimpleRulesRemover::remove(const grammar::LG < SymbolType > & origGrammar) {
	return removeNonEpsilonFree(origGrammar);
}

template < class SymbolType >
grammar::LeftLG < SymbolType > SimpleRulesRemover::remove(const grammar::LeftLG < SymbolType > & origGrammar) {
	return removeNonEpsilonFree(origGrammar);
}

template < class SymbolType >
grammar::LeftRG < SymbolType > SimpleRulesRemover::remove(const grammar::LeftRG < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::RightLG < SymbolType > SimpleRulesRemover::remove(const grammar::RightLG < SymbolType > & origGrammar) {
	return removeNonEpsilonFree(origGrammar);
}

template < class SymbolType >
grammar::RightRG < SymbolType > SimpleRulesRemover::remove(const grammar::RightRG < SymbolType > & origGrammar) {
	return origGrammar;
}

} /* namespace simplify */

} /* namespace grammar */

#endif /* SIMPLE_RULES_REMOVER_H_ */
