/*
 * Follow.h
 *
 *  Created on: 9. 6. 2015
 *	  Author: Tomas Pecka
 */

#ifndef FOLLOW_H_
#define FOLLOW_H_

#include <vector>
#include <set>
#include <variant>
#include <algorithm>
#include <iterator>

#include <exception/CommonException.h>

#include <grammar/Grammar.h>
#include <string/Epsilon.h>

#include "First.h"

namespace grammar {

namespace parsing {

class Follow {
	template < class T, class SymbolType >
	static void follow ( const T & grammar, ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > & followSet );

public:
	template < class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > follow ( const T & grammar );

	template < class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > follow ( const T & grammar, const SymbolType & nt );
};

template < class T, class SymbolType >
void Follow::follow ( const T & grammar, ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > & followSet ) {
	for ( const std::pair < const SymbolType, ext::set < ext::vector < SymbolType > > > & rule : grammar.getRawRules ( ) ) {
		const SymbolType & X = rule.first;

		for ( const ext::vector < SymbolType > & rhs : rule.second )
			 // every nt in rhs is Y
			for ( typename ext::vector < SymbolType >::const_iterator it = rhs.begin ( ); it != rhs.end ( ); it++ ) {
				const SymbolType & Y = * it;

				if ( !grammar.getNonterminalAlphabet ( ).count ( Y ) )
					continue;

				ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > firstBeta = First::first ( grammar, ext::vector < SymbolType > ( std::next ( it ), rhs.end ( ) ) );

				if ( firstBeta.count ( ext::variant < SymbolType, string::Epsilon < SymbolType > >::template from < string::Epsilon < SymbolType > > ( ) ) ) {
					firstBeta.erase ( ext::variant < SymbolType, string::Epsilon < SymbolType > >::template from < string::Epsilon < SymbolType > > ( ) );
					followSet[Y].insert ( followSet[X].begin ( ), followSet[X].end ( ) );
				}

				followSet[Y].insert ( firstBeta.begin ( ), firstBeta.end ( ) );
			}

	}
}

template < class T, class SymbolType >
ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > Follow::follow ( const T & grammar ) {
	/*
	 * 1. Follow(S) = { \varepsilon }
	 *    Follow(A) = {} forall A \in N, A \neq S
	 * 2. Forall p \in P:
	 *      if p == X -> \alpha Y \beta
	 *		Follow(Y) = Follow(Y) \cup (First(\beta) \setminus { \varepsilon})
	 *	if p == X -> \alpha Y \beta \wedge \varepsilon \in First(\beta)
	 *		Follow(Y) = Follow(Y) \cup Follow(X)
	 * 3. goto 2 if any follow set was changed in prev step.
	 */

	ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > followSet1;

	for ( const SymbolType & symb : grammar.getNonterminalAlphabet ( ) )
		followSet1[symb];

	followSet1[grammar.getInitialSymbol ( )] = { ext::variant < SymbolType, string::Epsilon < SymbolType > >::template from < string::Epsilon < SymbolType > > ( ) };

	ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > followSet2 = followSet1;

	do {
		follow ( grammar, followSet2 );

		if ( followSet1 == followSet2 )
			break;

		followSet1 = followSet2;
	} while ( true );

	return followSet1;
}

template < class T, class SymbolType >
ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > Follow::follow ( const T & grammar, const SymbolType & nt ) {
	if ( !grammar.getNonterminalAlphabet ( ).count ( nt ) )
		throw exception::CommonException ( "Follow: Given symbol is not nonterminal." );

	return follow ( grammar )[nt];
}

} /* namespace parsing */

} /* namespace grammar */

#endif /* FOLLOW_H_ */
