/*
 * Kruskal.cpp
 *
 *   Created on: Mar 26, 2016
 *       Author: Jan Broz
 */

#include "Kruskal.h"

#include <vector>
#include <unordered_map>
#include <algorithm>
#include <stdexcept>

#include "../datastructs/Components.h"
#include <exception/CommonException.h>
#include <registration/AlgoRegistration.hpp>

namespace graph {

namespace spanningtree {

/// wrapper for edges containing weight
struct WeightedEdge {
	UndirectedEdge edge;
	int weight;
	WeightedEdge( const UndirectedEdge & e, int w ) : edge(e), weight(w) {}
};

bool isWeightSmaller( const WeightedEdge & e1, const WeightedEdge & e2)
{
	return e1.weight < e2.weight;
}

/// converts edges to wrappers with assigned weight
static ext::vector<WeightedEdge> getEdges( const UndirectedGraph & graph )
{
	ext::vector<WeightedEdge> edges;
	int weight;

	for (const UndirectedEdge & edge : graph.getEdges()) {
		try {
			weight = graph.getEdgeValue( edge );
		} catch (const std::out_of_range & ex) { // when this edge has no value assigned
			throw exception::CommonException("Kruskal: an edge is missing a weight");
		}
		edges.emplace_back( edge, weight );
	}

	return edges;
}

static AdjacencyListUndirectedGraph kruskal_impl( const UndirectedGraph & graph )
{
	AdjacencyListUndirectedGraph res;  // spanning tree
	ext::vector<WeightedEdge> edges;   // edges with weights
	Components comps;                  // components of the emerging spanning tree

	for (const Node & node : graph.getNodes()) {
		res.addNode( node );
		comps.MakeSet( node );
	}

	edges = getEdges( graph );
	// sort edges by increasing weight
	std::sort( edges.begin(), edges.end(), isWeightSmaller );

	for (const WeightedEdge & wedge : edges) { // little easter egg here :D
		if (comps.Find( wedge.edge.getFirstNode() ) != comps.Find( wedge.edge.getSecondNode() )) { // if both nodes of the edge are in same set, it would create cycle
			comps.Union( wedge.edge.getFirstNode(), wedge.edge.getSecondNode() );
			res.addEdge( wedge.edge, wedge.weight );
		}
	}

	return res;
}

AdjacencyListUndirectedGraph Kruskal::kruskal( const UndirectedGraph & graph )
{
	return kruskal_impl( graph );
}

// auto KruskalUndirectedGraph = registration::OverloadRegister < Kruskal,  graph::AdjacencyListUndirectedGraph, graph::UndirectedGraph > ( Kruskal::kruskal );

} // namespace spanningtree

} // namespace graph
