/*
 * OneTapeDTM.h
 *
 *  Created on: Apr 24, 2013
 *      Author: Martin Zak
 *      Author: Jan Travnicek
 */

#ifndef ONE_TAPE_DTM_H_
#define ONE_TAPE_DTM_H_

#include <map>
#include <set>
#include <tuple>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonException.h"
#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../common/Shift.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace automaton {

class TapeAlphabet;
class InputAlphabet;
class BlankSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * One tape turing machine
 */
template<class SymbolType, class StateType >
class OneTapeDTM final : public AutomatonBase, public alib::Components < OneTapeDTM < SymbolType, StateType >, ext::set < SymbolType >, component::Set, std::tuple < TapeAlphabet, InputAlphabet >, SymbolType, component::Value, BlankSymbol, ext::set < StateType >, component::Set, std::tuple < States, FinalStates >, StateType, component::Value, InitialState > {
protected:
	ext::map < ext::pair < StateType, SymbolType >, ext::tuple < StateType, SymbolType, Shift > > transitions;

public:
	explicit OneTapeDTM ( ext::set < StateType > states, ext::set < SymbolType > tapeAlphabet, SymbolType blankSymbol, ext::set< SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates );

	explicit OneTapeDTM ( StateType initialState, SymbolType blank );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const & {
		return this-> template accessComponent < InitialState > ( ).get ( );
	}

	StateType && getInitialState ( ) && {
		return std::move ( this-> template accessComponent < InitialState > ( ).get ( ) );
	}

	bool setInitialState ( StateType state ) {
		return this-> template accessComponent < InitialState > ( ).set ( std::move ( state ) );
	}

	const ext::set < StateType > & getStates ( ) const & {
		return this-> template accessComponent < States > ( ).get ( );
	}

	ext::set < StateType > && getStates ( ) && {
		return std::move ( this-> template accessComponent < States > ( ).get ( ) );
	}

	bool addState ( StateType state ) {
		return this-> template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( ext::set < StateType > states ) {
		this-> template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this-> template accessComponent < States > ( ).remove ( state );
	}

	const ext::set < StateType > & getFinalStates ( ) const & {
		return this-> template accessComponent < FinalStates > ( ).get ( );
	}

	ext::set < StateType > && getFinalStates ( ) && {
		return std::move ( this-> template accessComponent < FinalStates > ( ).get ( ) );
	}

	bool addFinalState ( StateType state ) {
		return this-> template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( ext::set < StateType > states ) {
		this-> template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this-> template accessComponent < FinalStates > ( ).remove ( state );
	}

	const ext::set < SymbolType > & getInputAlphabet ( ) const & {
		return this-> template accessComponent < InputAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getInputAlphabet ( ) && {
		return std::move ( this-> template accessComponent < InputAlphabet > ( ).get ( ) );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this-> template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( ext::set < SymbolType > symbols ) {
		this-> template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( ext::set < SymbolType > symbols ) {
		this-> template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this-> template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	const ext::set < SymbolType > & getTapeAlphabet ( ) const & {
		return this-> template accessComponent < TapeAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getTapeAlphabet ( ) && {
		return std::move ( this-> template accessComponent < TapeAlphabet > ( ).get ( ) );
	}

	bool addTapeSymbol ( SymbolType symbol ) {
		return this-> template accessComponent < TapeAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addTapeSymbols ( ext::set < SymbolType > symbols ) {
		this-> template accessComponent < TapeAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setTapeAlphabet ( ext::set < SymbolType > symbols ) {
		this-> template accessComponent < TapeAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeTapeSymbol ( const SymbolType & symbol ) {
		this-> template accessComponent < TapeAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getBlankSymbol ( ) const & {
		return this-> template accessComponent < BlankSymbol > ( ).get ( );
	}

	SymbolType && getBlankSymbol ( ) && {
		return std::move ( this-> template accessComponent < BlankSymbol > ( ).get ( ) );
	}

	bool setBlankSymbol ( SymbolType state ) {
		return this-> template accessComponent < BlankSymbol > ( ).set ( std::move ( state ) );
	}

	/**
	 * Adds transition to the automaton.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the TM (state, tape symbol) or when transition already exists
	 */
	bool addTransition ( StateType from, SymbolType input, StateType to, SymbolType output, Shift shift );

	/**
	 * Removes the transition from the TM.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the TM
	 */
	bool removeTransition ( const StateType & from, const SymbolType & input, const StateType & to, const SymbolType & output, const Shift & shift );

	/**
	 * @return TM transitions
	 */
	const ext::map < ext::pair < StateType, SymbolType >, ext::tuple < StateType, SymbolType, Shift > > & getTransitions ( ) const &;

	ext::map < ext::pair < StateType, SymbolType >, ext::tuple < StateType, SymbolType, Shift > > && getTransitions ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const OneTapeDTM & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "OneTapeDTM";

		return xmlTagName;
	}

	static OneTapeDTM parse ( ext::deque < sax::Token >::iterator & input );
	static void parseTransition ( ext::deque < sax::Token >::iterator & input, OneTapeDTM & automaton );

	static void compose ( ext::deque < sax::Token > & out, const OneTapeDTM & automaton );
	static void composeTransitions ( ext::deque < sax::Token > & out, const OneTapeDTM & automaton );

	virtual alib::ObjectBase * inc ( ) &&;

	typedef OneTapeDTM < > normalized_type;
};

template<class SymbolType, class StateType >
OneTapeDTM<SymbolType, StateType>::OneTapeDTM ( ext::set < StateType > states, ext::set < SymbolType > tapeAlphabet, SymbolType blankSymbol, ext::set< SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates ) : alib::Components < OneTapeDTM, ext::set < SymbolType >, component::Set, std::tuple < TapeAlphabet, InputAlphabet >, SymbolType, component::Value, BlankSymbol, ext::set < StateType>, component::Set, std::tuple < States, FinalStates >, StateType, component::Value, InitialState > ( std::move ( tapeAlphabet ), std::move ( inputAlphabet ), std::move ( blankSymbol ), std::move ( states ), std::move ( finalStates ), std::move ( initialState ) ) {

}

template<class SymbolType, class StateType >
OneTapeDTM<SymbolType, StateType>::OneTapeDTM(StateType initial, SymbolType blank) : OneTapeDTM(ext::set<StateType>{initial}, ext::set<SymbolType> {blank}, blank, ext::set<SymbolType>{}, initial, ext::set<StateType>{}) {

}

template<class SymbolType, class StateType >
AutomatonBase* OneTapeDTM<SymbolType, StateType>::clone() const {
	return new OneTapeDTM<SymbolType, StateType>(*this);
}

template<class SymbolType, class StateType >
AutomatonBase* OneTapeDTM<SymbolType, StateType>::plunder() && {
	return new OneTapeDTM<SymbolType, StateType>(std::move(*this));
}

template<class SymbolType, class StateType >
bool OneTapeDTM<SymbolType, StateType>::addTransition(StateType from, SymbolType input, StateType to, SymbolType output, Shift shift) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + ext::to_string ( from ) + "\" doesn't exist.");
	}

	if (!getTapeAlphabet().count(input)) {
		throw AutomatonException("Tape symbol \"" + ext::to_string ( input ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + ext::to_string ( to ) + "\" doesn't exist.");
	}

	if (!getTapeAlphabet().count(output)) {
		throw AutomatonException("Tape symbol  \"" + ext::to_string ( output ) + "\" doesn't exist.");
	}

	ext::pair<StateType, SymbolType> key = ext::make_pair(std::move(from), std::move(input));

	ext::tuple<StateType, SymbolType, Shift > value(std::move(to), std::move(output), shift);

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else
			throw AutomatonException("Transition (\"" + ext::to_string ( key.first ) + "\", \"" + ext::to_string ( key.second ) + "\") -> ? already exists.");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)));
	return true;
}

template<class SymbolType, class StateType >
bool OneTapeDTM<SymbolType, StateType>::removeTransition(const StateType& from, const SymbolType& input, const StateType& to, const SymbolType& output, const Shift& shift) {
	ext::pair<StateType, SymbolType> key = ext::make_pair(from, input);

	if (transitions.find(key) == transitions.end())
		return false;

	ext::tuple<StateType, SymbolType, Shift > value(to, output, shift);
	if(transitions.find(key)->second != value) 
		throw AutomatonException("Transition (\"" + ext::to_string ( from ) + "\", \"" + ext::to_string ( input ) + "\") -> ? doesn't exists.");

	transitions.erase(key);
	return true;
}

template<class SymbolType, class StateType >
const ext::map < ext::pair < StateType, SymbolType >, ext::tuple < StateType, SymbolType, Shift > > & OneTapeDTM < SymbolType, StateType >::getTransitions ( ) const & {
	return transitions;
}

template<class SymbolType, class StateType >
ext::map < ext::pair < StateType, SymbolType >, ext::tuple < StateType, SymbolType, Shift > > && OneTapeDTM < SymbolType, StateType >::getTransitions ( ) && {
	return std::move ( transitions );
}

template<class SymbolType, class StateType >
int OneTapeDTM<SymbolType, StateType>::compare(const OneTapeDTM<SymbolType, StateType>& other) const {
	auto first = ext::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getTapeAlphabet(), getBlankSymbol(), transitions);
	auto second = ext::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getTapeAlphabet(), other.getBlankSymbol(), other.transitions);

	static ext::compare<decltype(first)> comp;
	return comp(first, second);
}

template<class SymbolType, class StateType >
void OneTapeDTM<SymbolType, StateType>::operator>>(std::ostream& out) const {
	out << "(OneTapeDTM "
			<< "states = " << getStates()
			<< "inputAlphabet = " << getInputAlphabet()
			<< "initialState = " << getInitialState()
			<< "finalStates = " << getFinalStates()
			<< "tapeAlphabet = " << getTapeAlphabet()
			<< "blankSymbol = " << getBlankSymbol()
			<< "transitions = " << transitions
			<< ")";
}

template<class SymbolType, class StateType >
OneTapeDTM<SymbolType, StateType>::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template<class SymbolType, class StateType >
OneTapeDTM<SymbolType, StateType> OneTapeDTM<SymbolType, StateType>::parse(ext::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, OneTapeDTM::getXmlTagName());

	ext::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	ext::set<SymbolType> tapeSymbols = AutomatonFromXMLParser::parseTapeAlphabet<SymbolType>(input);
	ext::set<SymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<SymbolType>(input);
	SymbolType blank = AutomatonFromXMLParser::parseBlankSymbol<SymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	ext::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	OneTapeDTM automaton(std::move(initialState), std::move(blank));
	automaton.setStates(std::move(states));
	automaton.setTapeAlphabet(std::move(tapeSymbols));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<OneTapeDTM>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, OneTapeDTM::getXmlTagName());
	return automaton;
}

template<class SymbolType, class StateType >
void OneTapeDTM<SymbolType, StateType>::parseTransition(ext::deque<sax::Token>::iterator& input, OneTapeDTM<SymbolType, StateType>& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
	SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<SymbolType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	SymbolType outputSymbol = AutomatonFromXMLParser::parseTransitionOutputSymbol<SymbolType>(input);
	Shift shift = AutomatonFromXMLParser::parseTransitionShift(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(outputSymbol), shift);
}

template<class SymbolType, class StateType >
void OneTapeDTM<SymbolType, StateType>::compose(ext::deque<sax::Token>& out, const OneTapeDTM & automaton ) {
	out.emplace_back(OneTapeDTM::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, automaton.getStates());
	AutomatonToXMLComposer::composeTapeAlphabet(out, automaton.getTapeAlphabet());
	AutomatonToXMLComposer::composeInputAlphabet(out, automaton.getInputAlphabet());
	AutomatonToXMLComposer::composeBlankSymbol(out, automaton.getBlankSymbol());
	AutomatonToXMLComposer::composeInitialState(out, automaton.getInitialState());
	AutomatonToXMLComposer::composeFinalStates(out, automaton.getFinalStates());
	composeTransitions(out, automaton);

	out.emplace_back(OneTapeDTM::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template<class SymbolType, class StateType >
void OneTapeDTM<SymbolType, StateType>::composeTransitions(ext::deque<sax::Token>& out, const OneTapeDTM & automaton ) {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : automaton.getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, std::get<0>(transition.second));
		AutomatonToXMLComposer::composeTransitionOutputSymbol(out, std::get<1>(transition.second));
		AutomatonToXMLComposer::composeTransitionShift(out, std::get<2>(transition.second));

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class StateType >
alib::ObjectBase* OneTapeDTM < SymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace alib {

template<class SymbolType, class StateType >
class SetConstraint< automaton::OneTapeDTM<SymbolType, StateType>, SymbolType, automaton::TapeAlphabet > {
public:
	static bool used ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const SymbolType & symbol ) {
		if ( automaton.getBlankSymbol ( ) == symbol )
			return true;

		if ( automaton.getInputAlphabet().count(symbol))
			return true;

		for ( const std::pair<const ext::pair<StateType, SymbolType>, ext::tuple<StateType, SymbolType, automaton::Shift> >& transition : automaton.getTransitions())
			if (symbol == transition.first.second || symbol == std::get<1>(transition.second))
				return true;

		return false;

	}

	static bool available ( const automaton::OneTapeDTM<SymbolType, StateType> &, const SymbolType &) {
		return true;
	}

	static void valid ( const automaton::OneTapeDTM<SymbolType, StateType> &, const SymbolType & ) {
	}
};

template<class SymbolType, class StateType >
class SetConstraint< automaton::OneTapeDTM<SymbolType, StateType>, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::OneTapeDTM<SymbolType, StateType> &, const SymbolType & ) {
		return false;
	}

	static bool available ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const SymbolType & symbol ) {
		return automaton.getTapeAlphabet ( ).count ( symbol );
	}

	static void valid ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const SymbolType & symbol ) {
		if (symbol == automaton.getBlankSymbol())
			throw automaton::AutomatonException("Input symbol \"" + ext::to_string ( symbol ) + "\" cannot be blank symbol.");
	}
};

template<class SymbolType, class StateType >
class ElementConstraint< automaton::OneTapeDTM<SymbolType, StateType>, SymbolType, automaton::BlankSymbol > {
public:
	static bool available ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const SymbolType & symbol ) {
		return automaton.getTapeAlphabet ( ).count ( symbol );
	}

	static void valid ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const SymbolType & symbol ) {
		if (automaton.getInputAlphabet().count( symbol ))
			throw automaton::AutomatonException("Blank symbol \"" + ext::to_string ( symbol ) + "\" cannot be in input alphabet.");
	}
};

template<class SymbolType, class StateType >
class SetConstraint< automaton::OneTapeDTM<SymbolType, StateType>, StateType, automaton::States > {
public:
	static bool used ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair<const ext::pair<StateType, SymbolType>, ext::tuple<StateType, SymbolType, automaton::Shift> >& transition : automaton.getTransitions ( ) )
			if ( state == transition.first.first || state == std::get < 0 > ( transition.second ) )
				return true;

		return false;
	}

	static bool available ( const automaton::OneTapeDTM<SymbolType, StateType> &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::OneTapeDTM<SymbolType, StateType> &, const StateType & ) {
	}
};

template<class SymbolType, class StateType >
class SetConstraint< automaton::OneTapeDTM<SymbolType, StateType>, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::OneTapeDTM<SymbolType, StateType> &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const StateType & state ) {
		return automaton.getStates ( ).count ( state );
	}

	static void valid ( const automaton::OneTapeDTM<SymbolType, StateType> &, const StateType & ) {
		//TODO final state and outgoing transitions
	}
};

template<class SymbolType, class StateType >
class ElementConstraint< automaton::OneTapeDTM<SymbolType, StateType>, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::OneTapeDTM<SymbolType, StateType> & automaton, const StateType & state ) {
		return automaton.getStates ( ).count ( state );
	}

	static void valid ( const automaton::OneTapeDTM<SymbolType, StateType> &, const StateType & ) {
	}
};

/**
 * Helper for normalisation of types specified by templates used as internal datatypes of symbols and states.
 *
 * \returns new instance of the automaton with default template parameters or unmodified instance if the template parameters were already the default ones
 */
template<class SymbolType, class StateType >
struct normalize < automaton::OneTapeDTM < SymbolType, StateType >, typename std::enable_if < ! std::is_same < automaton::OneTapeDTM < SymbolType, StateType >, automaton::OneTapeDTM < > >::value >::type > {
	static automaton::OneTapeDTM < > eval ( automaton::OneTapeDTM < SymbolType, StateType > && value ) {
		ext::set < DefaultSymbolType > tapeAlphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getTapeAlphabet ( ) );
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getInputAlphabet ( ) );
		DefaultSymbolType blankSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( value ).getBlankSymbol ( ) );
		ext::set < DefaultStateType > states = automaton::AutomatonNormalize::normalizeStates ( std::move ( value ).getStates ( ) );
		DefaultStateType initialState = automaton::AutomatonNormalize::normalizeState ( std::move ( value ).getInitialState ( ) );
		ext::set < DefaultStateType > finalStates = automaton::AutomatonNormalize::normalizeStates ( std::move ( value ).getFinalStates ( ) );

		automaton::OneTapeDTM < > res ( std::move ( states ), std::move ( tapeAlphabet ), std::move ( blankSymbol ), std::move ( alphabet ), std::move ( initialState ), std::move ( finalStates ) );

		for ( std::pair < ext::pair < DefaultStateType, DefaultSymbolType >, ext::tuple < DefaultStateType, DefaultSymbolType, automaton::Shift > > && transition : ext::make_moveable_map ( std::move ( value ).getTransitions ( ) ) ) {
			DefaultStateType from = automaton::AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			DefaultSymbolType input = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( transition.first.second ) );
			DefaultStateType to = automaton::AutomatonNormalize::normalizeState ( std::move ( std::get < 0 > ( transition.second ) ) );
			DefaultSymbolType output = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( std::get < 1 > ( transition.second ) ) );

			res.addTransition ( std::move ( from ), std::move ( input ), std::move ( to ), std::move ( output ), std::get < 2 > ( transition.second ) );
		}

		return res;
	}
};

} /* namespace alib */

#endif /* ONE_TAPE_DTM_H_ */
