/*
 * EpsilonFreeCFG.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_FREE_CFG_H_
#define EPSILON_FREE_CFG_H_

#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../GrammarBase.h"
#include "../GrammarFeatures.h"
#include "../GrammarException.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../common/GrammarNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace grammar {

/**
 * Epsilon free context free grammar. Type 2 in Chomsky hierarchy. Produces context free languages.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

template < class SymbolType >
class EpsilonFreeCFG final : public GrammarBase, public alib::Components < EpsilonFreeCFG < SymbolType >, ext::set < SymbolType >, component::Set, std::tuple < TerminalAlphabet, NonterminalAlphabet >, SymbolType, component::Value, InitialSymbol > {
	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > rules;
	bool generatesEpsilon;

public:
	explicit EpsilonFreeCFG ( SymbolType initialSymbol );

	explicit EpsilonFreeCFG ( ext::set < SymbolType > nonTerminalSymbols, ext::set < SymbolType > terminalSymbols, SymbolType initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide );

	void addRules ( SymbolType leftHandSide, ext::set < ext::vector < SymbolType > > rightHandSide );

	const ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > & getRules ( ) const &;

	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > && getRules ( ) &&;

	bool removeRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide );

	bool addRawRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide );

	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > getRawRules ( ) const;

	bool removeRawRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide );

	const SymbolType & getInitialSymbol ( ) const & {
		return this->template accessComponent < InitialSymbol > ( ).get ( );
	}

	SymbolType && getInitialSymbol ( ) && {
		return std::move ( this->template accessComponent < InitialSymbol > ( ).get ( ) );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const ext::set < SymbolType > & getNonterminalAlphabet ( ) const & {
		return this->template accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getNonterminalAlphabet ( ) && {
		return std::move ( this->template accessComponent < NonterminalAlphabet > ( ).get ( ) );
	}

	bool addNonterminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const ext::set < SymbolType > & getTerminalAlphabet ( ) const & {
		return this->template accessComponent < TerminalAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getTerminalAlphabet ( ) && {
		return std::move ( this->template accessComponent < TerminalAlphabet > ( ).get ( ) );
	}

	bool addTerminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void setGeneratesEpsilon ( bool genEps );
	bool getGeneratesEpsilon ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const EpsilonFreeCFG & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "EpsilonFreeCFG";

		return xmlTagName;
	}

	static EpsilonFreeCFG parse ( ext::deque < sax::Token >::iterator & input );
	static void parseRule ( ext::deque < sax::Token >::iterator & input, EpsilonFreeCFG & grammar );

	static void compose ( ext::deque < sax::Token > & out, const EpsilonFreeCFG & grammar );
	static void composeRules ( ext::deque < sax::Token > & out, const EpsilonFreeCFG & grammar );

	virtual alib::ObjectBase * inc ( ) &&;

	typedef EpsilonFreeCFG < > normalized_type;
};

template < class SymbolType >
EpsilonFreeCFG < SymbolType >::EpsilonFreeCFG ( SymbolType initialSymbol ) : EpsilonFreeCFG ( ext::set < SymbolType > { initialSymbol }, ext::set < SymbolType > ( ), initialSymbol ) {
}

template < class SymbolType >
EpsilonFreeCFG < SymbolType >::EpsilonFreeCFG ( ext::set < SymbolType > nonterminalAlphabet, ext::set < SymbolType > terminalAlphabet, SymbolType initialSymbol ) : alib::Components < EpsilonFreeCFG, ext::set < SymbolType >, component::Set, std::tuple < TerminalAlphabet, NonterminalAlphabet >, SymbolType, component::Value, InitialSymbol > ( std::move ( terminalAlphabet), std::move ( nonterminalAlphabet ), std::move ( initialSymbol ) ), generatesEpsilon ( false ) {
}

template < class SymbolType >
GrammarBase * EpsilonFreeCFG < SymbolType >::clone ( ) const {
	return new EpsilonFreeCFG ( * this );
}

template < class SymbolType >
GrammarBase * EpsilonFreeCFG < SymbolType >::plunder ( ) && {
	return new EpsilonFreeCFG ( std::move ( * this ) );
}

template < class SymbolType >
bool EpsilonFreeCFG < SymbolType >::addRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 )
		throw GrammarException ( "Epsilon rule is not allowed" );

	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const SymbolType & symbol : rightHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ext::to_string ( symbol ) + "\" is not neither terminal nor nonterminal symbol" );

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

template < class SymbolType >
void EpsilonFreeCFG < SymbolType >::addRules ( SymbolType leftHandSide, ext::set < ext::vector < SymbolType > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const ext::vector < SymbolType > & rhs : rightHandSide ) {
		if ( rhs.size ( ) == 0 )
			throw GrammarException ( "Epsilon rule is not allowed" );

		for ( const SymbolType & symbol : rhs )
			if ( ! getTerminalAlphabet ( ).count ( symbol ) && ! getNonterminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol \"" + ext::to_string ( symbol ) + "\" is not neither terminal nor nonterminal symbol" );
	}

	rules [ std::move ( leftHandSide ) ].insert ( ext::make_moveable_set ( rightHandSide ).begin ( ), ext::make_moveable_set ( rightHandSide ).end ( ) );
}

template < class SymbolType >
const ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > & EpsilonFreeCFG < SymbolType >::getRules ( ) const & {
	return rules;
}

template < class SymbolType >
ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > && EpsilonFreeCFG < SymbolType >::getRules ( ) && {
	return std::move ( rules );
}

template < class SymbolType >
bool EpsilonFreeCFG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

template < class SymbolType >
bool EpsilonFreeCFG < SymbolType >::addRawRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( true );
		return !res;
	} else {
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
	}
}

template < class SymbolType >
ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > EpsilonFreeCFG < SymbolType >::getRawRules ( ) const {
	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > res ( getRules ( ).begin ( ), getRules ( ).end ( ) );

	if ( getGeneratesEpsilon ( ) )
		res [ getInitialSymbol ( ) ].insert ( ext::vector < SymbolType > { } );

	return rules;
}

template < class SymbolType >
bool EpsilonFreeCFG < SymbolType >::removeRawRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( false );
		return res;
	} else {
		return removeRule ( leftHandSide, rightHandSide );
	}
}

template < class SymbolType >
void EpsilonFreeCFG < SymbolType >::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

template < class SymbolType >
bool EpsilonFreeCFG < SymbolType >::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

template < class SymbolType >
int EpsilonFreeCFG < SymbolType >::compare ( const EpsilonFreeCFG & other ) const {
	auto first = ext::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = ext::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void EpsilonFreeCFG < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(EpsilonFreeCFG "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules
	    << "generatesEpsilon = " << generatesEpsilon << ")";
}

template < class SymbolType >
EpsilonFreeCFG < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
EpsilonFreeCFG < SymbolType > EpsilonFreeCFG < SymbolType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, EpsilonFreeCFG::getXmlTagName() );

	ext::set < SymbolType > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet < SymbolType > ( input );
	ext::set < SymbolType > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet < SymbolType > ( input );
	SymbolType initialSymbol = GrammarFromXMLParser::parseInitialSymbol < SymbolType > ( input );

	EpsilonFreeCFG < SymbolType > grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );
	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, EpsilonFreeCFG::getXmlTagName() );
	return grammar;
}

template < class SymbolType >
void EpsilonFreeCFG < SymbolType >::parseRule ( ext::deque < sax::Token >::iterator & input, EpsilonFreeCFG & grammar ) {
	SymbolType lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS < SymbolType > ( input );
	ext::vector < SymbolType > rhs = GrammarFromXMLParser::parseRuleRHS < SymbolType > ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

template < class SymbolType >
void EpsilonFreeCFG < SymbolType >::compose ( ext::deque < sax::Token > & out, const EpsilonFreeCFG & grammar ) {
	out.emplace_back ( EpsilonFreeCFG::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, grammar.getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, grammar.getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, grammar.getInitialSymbol ( ) );
	composeRules ( out, grammar );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, grammar.getGeneratesEpsilon ( ) );

	out.emplace_back ( EpsilonFreeCFG::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
void EpsilonFreeCFG < SymbolType >::composeRules ( ext::deque < sax::Token > & out, const EpsilonFreeCFG & grammar ) {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : grammar.getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* EpsilonFreeCFG < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace grammar */

namespace alib {

template < class SymbolType >
class SetConstraint< grammar::EpsilonFreeCFG < SymbolType >, SymbolType, grammar::TerminalAlphabet > {
public:
	static bool used ( const grammar::EpsilonFreeCFG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::set < ext::vector < SymbolType > > > & rule : grammar.getRules ( ) )
			for ( const ext::vector < SymbolType > & rhs : rule.second )
				if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
					return true;

		return false;
	}

	static bool available ( const grammar::EpsilonFreeCFG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::EpsilonFreeCFG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
	}
};

template < class SymbolType >
class SetConstraint< grammar::EpsilonFreeCFG < SymbolType >, SymbolType, grammar::NonterminalAlphabet > {
public:
	static bool used ( const grammar::EpsilonFreeCFG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::set < ext::vector < SymbolType > > > & rule : grammar.getRules ( ) ) {
			if ( rule.first == symbol )
				return true;

			for ( const ext::vector < SymbolType > & rhs : rule.second )
				if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
					return true;

		}

		if ( grammar.template accessComponent < grammar::InitialSymbol > ( ).get ( ) == symbol )
			return true;

		return false;
	}

	static bool available ( const grammar::EpsilonFreeCFG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::EpsilonFreeCFG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
	}
};

template < class SymbolType >
class ElementConstraint< grammar::EpsilonFreeCFG < SymbolType >, SymbolType, grammar::InitialSymbol > {
public:
	static bool available ( const grammar::EpsilonFreeCFG < SymbolType > & grammar, const SymbolType & symbol ) {
		return grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const grammar::EpsilonFreeCFG < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < grammar::EpsilonFreeCFG < SymbolType >, typename std::enable_if < ! std::is_same < grammar::EpsilonFreeCFG < SymbolType >, grammar::EpsilonFreeCFG < > >::value >::type > {
	static grammar::EpsilonFreeCFG < > eval ( grammar::EpsilonFreeCFG < SymbolType > && value ) {
		ext::set < DefaultSymbolType > nonterminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getNonterminalAlphabet ( ) );
		ext::set < DefaultSymbolType > terminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getTerminalAlphabet ( ) );
		DefaultSymbolType initialSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( value ).getInitialSymbol ( ) );

		grammar::EpsilonFreeCFG < > res ( std::move ( nonterminals ), std::move ( terminals ), std::move ( initialSymbol ) );

		for ( std::pair < SymbolType, ext::set < ext::vector < SymbolType > > > && rule : ext::make_moveable_map ( std::move ( value ).getRules ( ) ) ) {

			ext::set < ext::vector < DefaultSymbolType > > rhs;
			for ( ext::vector < SymbolType > && target : ext::make_moveable_set ( rule.second ) )
				rhs.insert ( alphabet::SymbolNormalize::normalizeSymbols ( std::move ( target ) ) );

			DefaultSymbolType lhs = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( rule.first ) );

			res.addRules ( std::move ( lhs ), std::move ( rhs ) );
		}

		res.setGeneratesEpsilon ( value.getGeneratesEpsilon ( ) );

		return res;
	}
};

} /* namespace alib */

#endif /* EPSILON_FREE_CFG_H_ */
