/*
 * BNDMMatcher.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef BNDM_MATCHER_H_
#define BNDM_MATCHER_H_

#include <set>
#include <string>
#include <iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsMap.h>
#include <container/ObjectsVector.h>

#include <container/xml/ObjectsSet.h>
#include <container/xml/ObjectsMap.h>
#include <container/xml/ObjectsVector.h>
#include <container/xml/ObjectsBitset.h>

#include <primitive/Bool.h>
#include <primitive/xml/Bool.h>
#include <bitset>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType, size_t BitmaskBitCount = 64 >
class BNDMMatcher final : public alib::ObjectBase, public alib::Components < BNDMMatcher < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
protected:
	ext::map < SymbolType, ext::bitset < BitmaskBitCount > > m_vectors;
	ext::vector < SymbolType > m_string;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit BNDMMatcher ( ext::set < SymbolType > alphabet, ext::map < SymbolType, ext::bitset < BitmaskBitCount > > vectors, ext::vector < SymbolType > string );

	/**
	 * @return Root node of the trie
	 */
	const ext::map < SymbolType, ext::bitset < BitmaskBitCount > > & getData ( ) const &;

	ext::map < SymbolType, ext::bitset < BitmaskBitCount > > && getData ( ) &&;

	const ext::vector < SymbolType > & getString ( ) const &;

	ext::vector < SymbolType > && getString ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	/**
	 * Sets the bit vector for given symbol
	 * @param tree root node to set
	 */
	void setBitVectorForSymbol ( SymbolType symbol, ext::bitset < BitmaskBitCount > data );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const BNDMMatcher & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef BNDMMatcher < > normalized_type;
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType, size_t BitmaskBitCount >
BNDMMatcher < SymbolType, BitmaskBitCount >::BNDMMatcher ( ext::set < SymbolType > alphabet, ext::map < SymbolType, ext::bitset < BitmaskBitCount > > vectors, ext::vector < SymbolType > string ) : alib::Components < BNDMMatcher, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( alphabet ) ), m_vectors ( std::move ( vectors ) ), m_string ( std::move ( string ) ) {
}

template < class SymbolType, size_t BitmaskBitCount >
alib::ObjectBase * BNDMMatcher < SymbolType, BitmaskBitCount >::clone ( ) const {
	return new BNDMMatcher ( * this );
}

template < class SymbolType, size_t BitmaskBitCount >
alib::ObjectBase * BNDMMatcher < SymbolType, BitmaskBitCount >::plunder ( ) && {
	return new BNDMMatcher ( std::move ( * this ) );
}

template < class SymbolType, size_t BitmaskBitCount >
const ext::map < SymbolType, ext::bitset < BitmaskBitCount > > & BNDMMatcher < SymbolType, BitmaskBitCount >::getData ( ) const & {
	return m_vectors;
}

template < class SymbolType, size_t BitmaskBitCount >
ext::map < SymbolType, ext::bitset < BitmaskBitCount > > && BNDMMatcher < SymbolType, BitmaskBitCount >::getData ( ) && {
	return std::move ( m_vectors );
}

template < class SymbolType, size_t BitmaskBitCount >
const ext::vector < SymbolType > & BNDMMatcher < SymbolType, BitmaskBitCount >::getString ( ) const & {
	return m_string;
}

template < class SymbolType, size_t BitmaskBitCount >
ext::vector < SymbolType > && BNDMMatcher < SymbolType, BitmaskBitCount >::getString ( ) && {
	return std::move ( m_string );
}

template < class SymbolType, size_t BitmaskBitCount >
void BNDMMatcher < SymbolType, BitmaskBitCount >::setBitVectorForSymbol ( SymbolType symbol, ext::bitset < BitmaskBitCount > data ) {
	this->m_vectors [ symbol ] = std::move ( data );
}

template < class SymbolType, size_t BitmaskBitCount >
void BNDMMatcher < SymbolType, BitmaskBitCount >::operator >>( std::ostream & out ) const {
	out << "(BNDMMatcher " << this->m_vectors << ")";
}

template < class SymbolType, size_t BitmaskBitCount >
int BNDMMatcher < SymbolType, BitmaskBitCount >::compare ( const BNDMMatcher & other ) const {
	auto first = ext::tie ( getData ( ), getAlphabet ( ) );
	auto second = ext::tie ( other.getData ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, size_t BitmaskBitCount >
BNDMMatcher < SymbolType, BitmaskBitCount >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, size_t BitmaskBitCount >
alib::ObjectBase* BNDMMatcher < SymbolType, BitmaskBitCount >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace alib {

template < class SymbolType, size_t BitmaskBitCount >
class SetConstraint < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount >, SymbolType, indexes::stringology::GeneralAlphabet > {
public:
	static bool used ( const indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > & index, const SymbolType & symbol ) {
		const ext::map < SymbolType, ext::bitset < BitmaskBitCount > > & content = index.getData ( );
		return content.find( symbol ) != content.end();
	}

	static bool available ( const indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > &, const SymbolType & ) {
	}
};

template < class SymbolType, size_t BitmaskBitCount >
struct normalize < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount >, typename std::enable_if < ! std::is_same < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount >, indexes::stringology::BNDMMatcher < > >::value && std::is_same < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount >, indexes::stringology::BNDMMatcher < DefaultSymbolType, BitmaskBitCount > >::value >::type > {
	static indexes::stringology::BNDMMatcher < DefaultSymbolType, BitmaskBitCount > && eval ( indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > && value ) {
		return std::move ( value );
	}
};

template < class SymbolType, size_t BitmaskBitCount >
struct normalize < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount >, typename std::enable_if < ! std::is_same < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount >, indexes::stringology::BNDMMatcher < DefaultSymbolType, BitmaskBitCount > >::value && ! std::is_same < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount >, indexes::stringology::BNDMMatcher < DefaultSymbolType, BitmaskBitCount > >::value >::type > {
	static indexes::stringology::BNDMMatcher < DefaultSymbolType, BitmaskBitCount > eval ( indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );

		ext::map < DefaultSymbolType, ext::bitset < BitmaskBitCount > > vectors;
		for ( std::pair < SymbolType, ext::bitset < BitmaskBitCount > > && vector : ext::make_moveable_map ( std::move ( value ).getData ( ) ) )
			vectors.insert ( std::make_pair ( alphabet::SymbolNormalize::normalizeSymbol ( std::move ( vector.first ) ), std::move ( vector.second ) ) );

		ext::vector < DefaultSymbolType > string = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( value ).getString ( ) );

		return indexes::stringology::BNDMMatcher < DefaultSymbolType, BitmaskBitCount > ( std::move ( alphabet ), std::move ( vectors ), std::move ( string ) );
	}
};

template < class SymbolType, size_t BitmaskBitCount >
struct xmlApi < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > > {
	static indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > & data );
};

template < class SymbolType, size_t BitmaskBitCount >
indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > xmlApi < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	ext::set < SymbolType > alphabet = alib::xmlApi < ext::set < SymbolType > >::parse ( input );
	ext::map < SymbolType, ext::bitset < BitmaskBitCount > > data = alib::xmlApi < ext::map < SymbolType, ext::bitset < BitmaskBitCount > > >::parse ( input );
	ext::vector < SymbolType > string = alib::xmlApi < ext::vector < SymbolType > >::parse ( input );
	indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > res ( std::move ( alphabet ), std::move ( data ), std::move ( string ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType, size_t BitmaskBitCount >
bool xmlApi < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType, size_t BitmaskBitCount >
const std::string & xmlApi < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > >::xmlTagName ( ) {
	static std::string xmlTagName = "BNDMMatcher";

	return xmlTagName;
}

template < class SymbolType, size_t BitmaskBitCount >
void xmlApi < indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > >::compose ( ext::deque < sax::Token > & output, const indexes::stringology::BNDMMatcher < SymbolType, BitmaskBitCount > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < ext::set < SymbolType > >::compose ( output, index.getAlphabet ( ) );
	alib::xmlApi < ext::map < SymbolType, ext::bitset < BitmaskBitCount > > >::compose ( output, index.getData ( ) );
	alib::xmlApi < ext::vector < SymbolType > >::compose ( output, index.getString ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace alib */

#endif /* BNDM_MATCHER_H_ */
