/*
 * CompressedBitParallelIndex.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef COMPRESSED_BIT_PARALLEL_INDEX_H_
#define COMPRESSED_BIT_PARALLEL_INDEX_H_

#include <set>
#include <string>
#include <iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsMap.h>

#include <container/xml/ObjectsSet.h>
#include <container/xml/ObjectsMap.h>

#include <common/SparseBoolVector.hpp>

#include <primitive/Bool.h>
#include <primitive/xml/Bool.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class CompressedBitParallelIndex final : public alib::ObjectBase, public alib::Components < CompressedBitParallelIndex < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
protected:
	ext::map < SymbolType, common::SparseBoolVector > m_vectors;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit CompressedBitParallelIndex ( ext::set < SymbolType > alphabet, ext::map < SymbolType, common::SparseBoolVector > vectors );

	/**
	 * @return Root node of the trie
	 */
	const ext::map < SymbolType, common::SparseBoolVector > & getData ( ) const &;

	ext::map < SymbolType, common::SparseBoolVector > && getData ( ) &&;

	ext::vector < SymbolType > getString ( ) const;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	/**
	 * Sets the compressedBit vector for given symbol
	 * @param tree root node to set
	 */
	void setCompressedBitVectorForSymbol ( SymbolType symbol, common::SparseBoolVector data );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const CompressedBitParallelIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef CompressedBitParallelIndex < > normalized_type;
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
CompressedBitParallelIndex < SymbolType >::CompressedBitParallelIndex ( ext::set < SymbolType > alphabet, ext::map < SymbolType, common::SparseBoolVector > vectors ) : alib::Components < CompressedBitParallelIndex, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( alphabet ) ), m_vectors ( std::move ( vectors ) ) {
}

template < class SymbolType >
alib::ObjectBase * CompressedBitParallelIndex < SymbolType >::clone ( ) const {
	return new CompressedBitParallelIndex ( * this );
}

template < class SymbolType >
alib::ObjectBase * CompressedBitParallelIndex < SymbolType >::plunder ( ) && {
	return new CompressedBitParallelIndex ( std::move ( * this ) );
}

template < class SymbolType >
const ext::map < SymbolType, common::SparseBoolVector > & CompressedBitParallelIndex < SymbolType >::getData ( ) const & {
	return m_vectors;
}

template < class SymbolType >
ext::map < SymbolType, common::SparseBoolVector > && CompressedBitParallelIndex < SymbolType >::getData ( ) && {
	return std::move ( m_vectors );
}

template < class SymbolType >
ext::vector < SymbolType > CompressedBitParallelIndex < SymbolType >::getString ( ) const {
	ext::vector < SymbolType > res;

	unsigned index = 0;
	do {
		for ( const std::pair < const SymbolType, common::SparseBoolVector > & compressedBitVector : m_vectors )
			if ( compressedBitVector.second.size ( ) > index && compressedBitVector.second [ index ] ) {
				res.push_back ( compressedBitVector.first );
				continue;
			}

	} while ( res.size ( ) == index ++ + 1 );

	return res;
}

template < class SymbolType >
void CompressedBitParallelIndex < SymbolType >::setCompressedBitVectorForSymbol ( SymbolType symbol, common::SparseBoolVector data ) {
	this->m_vectors [ symbol ] = std::move ( data );
}

template < class SymbolType >
void CompressedBitParallelIndex < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(CompressedBitParallelIndex " << this->m_vectors << ")";
}

template < class SymbolType >
int CompressedBitParallelIndex < SymbolType >::compare ( const CompressedBitParallelIndex & other ) const {
	auto first = ext::tie ( getData ( ), getAlphabet ( ) );
	auto second = ext::tie ( other.getData ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
CompressedBitParallelIndex < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
alib::ObjectBase* CompressedBitParallelIndex < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace alib {

template < class SymbolType >
class SetConstraint < indexes::stringology::CompressedBitParallelIndex < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {
public:
	static bool used ( const indexes::stringology::CompressedBitParallelIndex < SymbolType > & index, const SymbolType & symbol ) {
		const ext::map < SymbolType, common::SparseBoolVector > & content = index.getData ( );
		return content.find( symbol ) != content.end();
	}

	static bool available ( const indexes::stringology::CompressedBitParallelIndex < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::CompressedBitParallelIndex < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < indexes::stringology::CompressedBitParallelIndex < SymbolType >, typename std::enable_if < ! std::is_same < indexes::stringology::CompressedBitParallelIndex < SymbolType >, indexes::stringology::CompressedBitParallelIndex < > >::value >::type > {
	static indexes::stringology::CompressedBitParallelIndex < > eval ( indexes::stringology::CompressedBitParallelIndex < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );

		ext::map < DefaultSymbolType, common::SparseBoolVector > vectors;
		for ( std::pair < SymbolType, common::SparseBoolVector > && vector : ext::make_moveable_map ( std::move ( value ).getData ( ) ) )
			vectors.insert ( std::make_pair ( alphabet::SymbolNormalize::normalizeSymbol ( std::move ( vector.first ) ), std::move ( vector.second ) ) );

		return indexes::stringology::CompressedBitParallelIndex < > ( std::move ( alphabet ), std::move ( vectors ) );
	}
};

template < class SymbolType >
struct xmlApi < indexes::stringology::CompressedBitParallelIndex < SymbolType > > {
	static indexes::stringology::CompressedBitParallelIndex < SymbolType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::stringology::CompressedBitParallelIndex < SymbolType > & data );
};

template < class SymbolType >
indexes::stringology::CompressedBitParallelIndex < SymbolType > xmlApi < indexes::stringology::CompressedBitParallelIndex < SymbolType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	ext::set < SymbolType > alphabet = alib::xmlApi < ext::set < SymbolType > >::parse ( input );
	ext::map < SymbolType, common::SparseBoolVector > data = alib::xmlApi < ext::map < SymbolType, common::SparseBoolVector > >::parse ( input );
	indexes::stringology::CompressedBitParallelIndex < SymbolType > res ( std::move ( alphabet ), std::move ( data ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType >
bool xmlApi < indexes::stringology::CompressedBitParallelIndex < SymbolType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType >
const std::string & xmlApi < indexes::stringology::CompressedBitParallelIndex < SymbolType > >::xmlTagName ( ) {
	static std::string xmlTagName = "CompressedBitParallelIndex";

	return xmlTagName;
}

template < class SymbolType >
void xmlApi < indexes::stringology::CompressedBitParallelIndex < SymbolType > >::compose ( ext::deque < sax::Token > & output, const indexes::stringology::CompressedBitParallelIndex < SymbolType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < ext::set < SymbolType > >::compose ( output, index.getAlphabet ( ) );
	alib::xmlApi < ext::map < SymbolType, common::SparseBoolVector > >::compose ( output, index.getData ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace alib */

#endif /* COMPRESSED_BIT_PARALLEL_INDEX_H_ */
