/*
 * SuffixTrie.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef SUFFIX_TRIE_H_
#define SUFFIX_TRIE_H_

#include <string>
#include <set>
#include <trie>
#include <variant>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/Object.h>
#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsTrie.h>

#include <container/xml/ObjectsSet.h>
#include <container/xml/ObjectsVariant.h>
#include <container/xml/ObjectsTrie.h>
#include <object/Void.h>
#include <object/xml/Void.h>

#include <primitive/Unsigned.h>
#include <primitive/xml/Unsigned.h>

#include <alphabet/common/SymbolNormalize.h>
#include <indexes/common/IndexesNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class SuffixTrie final : public alib::ObjectBase, public alib::Components < SuffixTrie < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
protected:
	ext::trie < SymbolType, ext::variant < void, unsigned > > m_trie;

public:
	/**
	 * @copydoc SuffixTrie::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrie::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit SuffixTrie ( ext::set < SymbolType > edgeAlphabet );
	explicit SuffixTrie ( ext::set < SymbolType > edgeAlphabet, ext::trie < SymbolType, ext::variant < void, unsigned > > trie );
	explicit SuffixTrie ( ext::trie < SymbolType, ext::variant < void, unsigned > > trie );

	void checkTrie ( const ext::trie < SymbolType, ext::variant < void, unsigned > > & trie );

	/**
	 * @return Root node of the trie
	 */
	const ext::trie < SymbolType, ext::variant < void, unsigned > > & getRoot ( ) const &;

	ext::trie < SymbolType, ext::variant < void, unsigned > > && getRoot ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( ext::trie < SymbolType, ext::variant < void, unsigned > > tree );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromEdgeAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SuffixTrie & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef SuffixTrie < > normalized_type;
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
SuffixTrie < SymbolType >::SuffixTrie ( ext::set < SymbolType > edgeAlphabet ) : SuffixTrie ( std::move ( edgeAlphabet ), ext::trie < SymbolType, ext::variant < void, unsigned > > ( ext::variant < void, unsigned >::template from < void > ( ) ) ) {
}

template < class SymbolType >
SuffixTrie < SymbolType >::SuffixTrie ( ext::set < SymbolType > edgeAlphabet, ext::trie < SymbolType, ext::variant < void, unsigned > > trie ) : alib::Components < SuffixTrie, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( edgeAlphabet ) ), m_trie ( std::move ( trie ) ) {
	checkTrie ( this->m_trie );
}

template < class SymbolType >
SuffixTrie < SymbolType >::SuffixTrie ( ext::trie < SymbolType, ext::variant < void, unsigned > > trie ) : SuffixTrie ( computeMinimalEdgeAlphabet ( trie ), trie ) {
}

template < class SymbolType >
alib::ObjectBase * SuffixTrie < SymbolType >::clone ( ) const {
	return new SuffixTrie ( * this );
}

template < class SymbolType >
alib::ObjectBase * SuffixTrie < SymbolType >::plunder ( ) && {
	return new SuffixTrie ( std::move ( * this ) );
}

template < class SymbolType >
void SuffixTrie < SymbolType >::checkTrie ( const ext::trie < SymbolType, ext::variant < void, unsigned > > & trie ) {
	for ( const std::pair < const SymbolType, ext::trie < SymbolType, ext::variant < void, unsigned > > > & child : trie.getChildren ( ) ) {
		if ( ! getAlphabet ( ).count ( child.first ) )
			throw exception::CommonException ( "Symbol " + ext::to_string ( child.first ) + "not in the alphabet." );
		checkTrie ( child.second );
	}
}

template < class SymbolType >
const ext::trie < SymbolType, ext::variant < void, unsigned > > & SuffixTrie < SymbolType >::getRoot ( ) const & {
	return m_trie;
}

template < class SymbolType >
ext::trie < SymbolType, ext::variant < void, unsigned > > && SuffixTrie < SymbolType >::getRoot ( ) && {
	return std::move ( m_trie );
}

template < class SymbolType >
void SuffixTrie < SymbolType >::setTree ( ext::trie < SymbolType, ext::variant < void, unsigned > > trie ) {
	checkTrie ( trie );
	this->m_trie = std::move ( trie ).plunder ( );
}

template < class SymbolType >
void SuffixTrie < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(SuffixTrie " << this->m_trie << ")";
}

template < class SymbolType >
int SuffixTrie < SymbolType >::compare ( const SuffixTrie & other ) const {
	auto first = ext::tie ( getRoot ( ), getAlphabet ( ) );
	auto second = ext::tie ( other.getRoot ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
SuffixTrie < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
alib::ObjectBase* SuffixTrie < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace alib {

template < class SymbolType >
class SetConstraint < indexes::stringology::SuffixTrie < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {

	static bool used ( const ext::trie < SymbolType, ext::variant < void, unsigned > > & trie, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::trie < SymbolType, ext::variant < void, unsigned > > > & child : trie.getChildren ( ) ) {
			if ( symbol == child.first || checkTrie ( trie, child.second ) )
				return true;
		}
		return false;
	}

public:
	static bool used ( const indexes::stringology::SuffixTrie < SymbolType > & index, const SymbolType & symbol ) {
		return used ( index.getRoot ( ), symbol );
	}

	static bool available ( const indexes::stringology::SuffixTrie < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::SuffixTrie < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < indexes::stringology::SuffixTrie < SymbolType >, typename std::enable_if < ! std::is_same < indexes::stringology::SuffixTrie < SymbolType >, indexes::stringology::SuffixTrie < > >::value >::type > {
	static indexes::stringology::SuffixTrie < > eval ( indexes::stringology::SuffixTrie < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlpahet ( ) );
		ext::trie < DefaultSymbolType, ext::variant < void, unsigned > > trie = indexes::IndexesNormalize::normalizeTrie ( std::move ( value ).getRoot ( ) );

		return indexes::stringology::SuffixTrie < > ( std::move ( alphabet ), std::move ( trie ) );
	}
};

template < class SymbolType >
struct xmlApi < indexes::stringology::SuffixTrie < SymbolType > > {
	static indexes::stringology::SuffixTrie < SymbolType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::stringology::SuffixTrie < SymbolType > & data );
};

template < class SymbolType >
indexes::stringology::SuffixTrie < SymbolType > xmlApi < indexes::stringology::SuffixTrie < SymbolType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	ext::set < SymbolType > edgeAlphabet = alib::xmlApi < ext::set < SymbolType > >::parse ( input );
	ext::trie < SymbolType, ext::variant < void, unsigned > > root = alib::xmlApi < ext::trie < SymbolType, ext::variant < void, unsigned > > >::parse ( input );
	indexes::stringology::SuffixTrie < SymbolType > res ( std::move ( edgeAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType >
bool xmlApi < indexes::stringology::SuffixTrie < SymbolType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType >
const std::string & xmlApi < indexes::stringology::SuffixTrie < SymbolType > >::xmlTagName ( ) {
	static std::string xmlTagName = "SuffixTrie";

	return xmlTagName;
}

template < class SymbolType >
void xmlApi < indexes::stringology::SuffixTrie < SymbolType > >::compose ( ext::deque < sax::Token > & output, const indexes::stringology::SuffixTrie < SymbolType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < ext::set < SymbolType > >::compose ( output, index.getAlphabet ( ) );
	alib::xmlApi < ext::trie < SymbolType, ext::variant < void, unsigned > > >::compose ( output, index.getRoot ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace alib */

#endif /* SUFFIX_TRIE_H_ */
