/*
 * LinearStringTerminatingSymbol.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef LINEAR_STRING_TERMINATING_SYMBOL_H_
#define LINEAR_STRING_TERMINATING_SYMBOL_H_

#include <iostream>
#include <set>
#include <vector>

#include <string/StringBase.h>
#include <string/StringFeatures.h>

#include <alphabet/Symbol.h>
#include <core/components.hpp>

namespace string {

class GeneralAlphabet;
class TerminatingSymbol;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
class LinearStringTerminatingSymbol final : public StringBase, public alib::Components < LinearStringTerminatingSymbol, ext::set < DefaultSymbolType >, component::Set, GeneralAlphabet, DefaultSymbolType, component::Value, TerminatingSymbol > {
	ext::vector < DefaultSymbolType > m_Data;

public:
	explicit LinearStringTerminatingSymbol ( DefaultSymbolType terminatingSymbol );
	explicit LinearStringTerminatingSymbol ( ext::set < DefaultSymbolType > alphabet, DefaultSymbolType terminatingSymbol, ext::vector < DefaultSymbolType > data );
	explicit LinearStringTerminatingSymbol ( DefaultSymbolType terminatingSymbol, ext::vector < DefaultSymbolType > data );
	explicit LinearStringTerminatingSymbol ( DefaultSymbolType terminatingSymbol, const LinearString < > & str );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	virtual const ext::set < DefaultSymbolType > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const ext::set < DefaultSymbolType > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add( symbols );
	}

	const DefaultSymbolType & getTerminatingSymbol ( ) const {
		return accessComponent < TerminatingSymbol > ( ).get ( );
	}

	/**
	 * @param element to append
	 */
	void appendSymbol ( DefaultSymbolType symbol );

	/**
	 * @return List of symbols forming string (const version).
	 */
	const ext::vector < DefaultSymbolType > & getContent ( ) const;

	void setContent ( ext::vector < DefaultSymbolType > data );

	/**
	 * @return true if string is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const LinearStringTerminatingSymbol & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "LinearStringTerminatingSymbol";

		return xmlTagName;
	}

	static LinearStringTerminatingSymbol parse ( ext::deque < sax::Token >::iterator & input );

	static void compose ( ext::deque < sax::Token > & out, const LinearStringTerminatingSymbol & string );

	virtual alib::ObjectBase * inc ( ) &&;

	typedef LinearStringTerminatingSymbol normalized_type;
};

} /* namespace string */

namespace alib {

template < >
class SetConstraint< ::string::LinearStringTerminatingSymbol, DefaultSymbolType, ::string::GeneralAlphabet > {
public:
	static bool used ( const ::string::LinearStringTerminatingSymbol & str, const DefaultSymbolType & symbol ) {
		const ext::vector<DefaultSymbolType>& content = str.getContent ( );
		return std::find(content.begin(), content.end(), symbol) != content.end();
	}

	static bool available ( const ::string::LinearStringTerminatingSymbol &, const DefaultSymbolType & ) {
		return true;
	}

	static void valid ( const ::string::LinearStringTerminatingSymbol &, const DefaultSymbolType & ) {
	}
};

template < >
class ElementConstraint< ::string::LinearStringTerminatingSymbol, DefaultSymbolType, ::string::TerminatingSymbol > {
public:
	static bool available ( const ::string::LinearStringTerminatingSymbol & str, const DefaultSymbolType & symbol ) {
		return str.getAlphabet ( ).count ( symbol );
	}

	static void valid ( const ::string::LinearStringTerminatingSymbol &, const DefaultSymbolType & ) {
	}
};

} /* namespace alib */

#endif /* LINEAR_STRING_TERMINATING_SYMBOL_H_ */
