/*
 * arun.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>

#include <factory/XmlDataFactory.hpp>

int main(int argc, char* argv[]) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd("Automaton run binary", ' ', "0.01");

		std::vector<std::string> allowed;
		allowed.push_back("occurrences");
		allowed.push_back("accept");
		allowed.push_back("result");
		allowed.push_back("translate");
		TCLAP::ValuesConstraint<std::string> allowedVals( allowed );

		TCLAP::ValueArg<std::string> type(	"t",	"type",		"Run type",		false,	"accept",	&allowedVals);
		cmd.add( type );

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Input from file",	false,	"-",		"file");
		cmd.add( input );

		TCLAP::ValueArg<std::string> failState(	"f",	"fail_state",	"Override default fail state",	false,	"-",	"file");
		cmd.add( failState );

		TCLAP::ValueArg<std::string> automaton(	"a",	"automaton",	"Automaton to run",	false,	"-",		"file");
		cmd.add( automaton );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		cli::Environment environment;
		environment.setBinding ( "inputData", input.getValue ( ) );
		environment.setBinding ( "inputAutomaton", automaton.getValue ( ) );
		environment.setBinding ( "inputFailState", failState.getValue ( ) );
		environment.setBinding ( "stdout", "-" );

		measurements::start("Overal", measurements::Type::OVERALL);
		measurements::start("Input read", measurements::Type::AUXILIARY);

		cli::Parser parser ( cli::Lexer ( "execute <#inputData > $data" ) );
		parser.parse ( )->run ( environment );
		parser = cli::Parser ( cli::Lexer ( "execute <#inputAutomaton > $automaton" ) );
		parser.parse ( )->run ( environment );

		if( type.getValue() == "result" && failState.isSet ( ) ) {
			parser = cli::Parser ( cli::Lexer ( "execute <:DefaultStateType #inputFailState > $failState" ) );
			parser.parse ( )->run ( environment );
		}

		measurements::end();
		measurements::start("Algorithm", measurements::Type::MAIN);

		std::string cliCommand;
		if( type.getValue() == "occurrences") {
			cliCommand = "execute automaton::run::Occurrences $automaton $data > $output";
		} else if( type.getValue() == "accept") {
			cliCommand = "execute automaton::run::Accept $automaton $data > $output";
		} else if( type.getValue() == "result" && failState.isSet ( ) ) {
			cliCommand = "execute automaton::run::Result $automaton $data $failState > $output";
		} else if( type.getValue() == "result" ) {
			cliCommand = "execute automaton::run::Result $automaton $data > $output";
		} else if( type.getValue() == "translate") {
			cliCommand = "execute automaton::run::Translate $automaton $data > $output";
		} else {
			throw exception::CommonException( "Invalid run type" );
		}

		parser = cli::Parser ( cli::Lexer ( cliCommand ) );
		parser.parse ( )->run ( environment );

		measurements::end ( );
		measurements::start ( "Output write", measurements::Type::AUXILIARY );

		parser = cli::Parser ( cli::Lexer ( "execute $output >#stdout" ) );
		parser.parse ( )->run ( environment );

		measurements::end();
		measurements::end();

		if(measure.getValue()) common::Streams::measure << measurements::results() << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch (const TCLAP::ArgException& exception) {
		common::Streams::out << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		common::Streams::err << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch (...) {
		common::Streams::err << "Unknown exception caught." << std::endl;
		return 127;
	}
}
